# Main kernel code for the Monte Carlo simulation
#--------------------------------------------

import numpy as np
import pycuda.autoinit
import pycuda.driver as drv
from numpy import random as rd
from pycuda.compiler import SourceModule

#PREPROCESSORS
def PREC_N(n,box,size):
    '''
    Returns the list of neighbours for a given site in the lattice
    '''
    NLIST = np.zeros(size)
    for i in range(4):
        for j in range(4):
            NLIST[i+4*j] = n + box*(i%4) + j*box*size
    return NLIST.astype(np.int32)


def FM_N(grid, size):
    for i in range(size*size):
        grid[i*3+0] = 0.0
        grid[i*3+1] = 0.0
        grid[i*3+2] = 1.0

def AFM_N(grid, size):
    for i in range(size*size):
        u,v = np.random.random(),np.random.random()
        theta, phi = 2*np.pi*u, np.arccos(2*v-1)
        grid[i*3+0] = np.sin(phi)*np.cos(theta)
        grid[i*3+1] = np.sin(phi)*np.sin(theta)
        grid[i*3+2] = np.cos(phi)

def alt_FM_N(grid, s1, s2, size):
    for i in range(size*size):
        grid[i*4+0] = 0.0
        grid[i*4+1] = 0.0
        grid[i*4+2] = s1*(grid[i*4+3]==1) + s2*(grid[i*4+3]==2)

def alt_AFM_N(grid, s1, s2, size):
    for i in range(size*size):
        u,v = np.random.random(),np.random.random()
        theta, phi = 2*np.pi*u, np.arccos(2*v-1)
        grid[i*4+0] = np.sin(phi)*np.cos(theta)*(s1*(grid[i*4+3]) + s2*(1-grid[i*4+3]))
        grid[i*4+1] = np.sin(phi)*np.sin(theta)*(s1*(grid[i*4+3]) + s2*(1-grid[i*4+3]))
        grid[i*4+2] = np.cos(phi)*(s1*(grid[i*4+3]) + s2*(1-grid[i*4+3]))

#KERNEL CODE
'''
The kernel code is written in CUDA C and is compiled using the SourceModule class from the pycuda.compiler module.
The functions are called seperately from below the kernel code.
'''
# =============================================================================
dev_hamiltonian = SourceModule("""
//cuda
#include <curand.h>
#include <cuda_runtime.h>
//Lattice Utility
__global__ void cp_grid(float_t* grid, float_t* tf)
{
    int idx = blockIdx.x;
    int threadID = idx;
    grid[int(tf[threadID*4])*3] = tf[threadID*4+1];
    grid[int(tf[threadID*4])*3+1] = tf[threadID*4+2];
    grid[int(tf[threadID*4])*3+2] = tf[threadID*4+3];
}
__global__ void alt_cp_grid(float_t* grid, float_t* tf)
{
    int idx = blockIdx.x;
    int threadID = idx;
    grid[int(tf[threadID*4])*4] = tf[threadID*4+1];
    grid[int(tf[threadID*4])*4+1] = tf[threadID*4+2];
    grid[int(tf[threadID*4])*4+2] = tf[threadID*4+3];
}
//Vector preprocessing
__global__ void uvec_processor(float_t* u, float_t* v, float_t* s1, float_t* s2, float_t* s3, float_t* spin)
{
    int idx = blockIdx.x;
    float_t phi = 2.0*3.14159265359*u[idx];
    float_t theta = acosf(2.0*v[idx] - 1.0);
    s1[idx] = spin[0]*sinf(theta)*cosf(phi);
    s2[idx] = spin[0]*sinf(theta)*sinf(phi);
    s3[idx] = spin[0]*cosf(theta);
}

__global__ void alt_uvec_ising(float_t* u, float_t* v, float_t* s1, float_t* s2, float_t* s3, float_t* spin)
{
    int idx = blockIdx.x;
    s1[idx] = 0.0;
    s2[idx] = 0.0;
    if(v[idx] <= 0.5)
    {
        s3[idx] = spin[0];
    }
    else
    {
        s3[idx] = -1.0*spin[0];
    }
}

__global__ void NList_processor(float_t* nlist, int* res, int* __SIZE)
{
    int Idx = blockIdx.x;
    res[Idx] = __float2uint_ru(__SIZE[0]*__SIZE[0]*nlist[Idx]);
}

//Neighbour mapping
__device__ void N1_6_6_6_12(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row-1)*size + col-1 + size*size)%(size*size);
    NLIST[1] = (row*size     + col+1 + size*size)%(size*size);
    NLIST[2] = ((row-1)*size + col   + size*size)%(size*size);
    NLIST[3] = ((row)*size   + col-1 + size*size)%(size*size);
    NLIST[4] = ((row+1)*size + col+1 + size*size)%(size*size);
    NLIST[5] = ((row+1)*size + col   + size*size)%(size*size);
}

__device__ void N2_6_6_6_12(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row-2)*size + col-1 + size*size)%(size*size);
    NLIST[1] = ((row-1)*size + col+1 + size*size)%(size*size);
    NLIST[2] = ((row+1)*size + col-1 + size*size)%(size*size);
    NLIST[3] = ((row-1)*size + col-2 + size*size)%(size*size);
    NLIST[4] = ((row+1)*size + col+1 + size*size)%(size*size);
    NLIST[5] = ((row+1)*size + col+2 + size*size)%(size*size);
}

__device__ void N3_6_6_6_12(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row-2)*size + col-2 + size*size)%(size*size);
    NLIST[1] = ((row)*size   + col+2 + size*size)%(size*size);
    NLIST[2] = ((row-2)*size + col   + size*size)%(size*size);
    NLIST[3] = ((row)*size   + col-2 + size*size)%(size*size);
    NLIST[4] = ((row+2)*size + col+2 + size*size)%(size*size);
    NLIST[5] = ((row+2)*size + col   + size*size)%(size*size);
}

__device__ void N4_6_6_6_12(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row-2)*size + col-3 + size*size)%(size*size);
    NLIST[1] = ((row-3)*size + col-1 + size*size)%(size*size);
    NLIST[2] = ((row-3)*size + col-2 + size*size)%(size*size);
    NLIST[3] = ((row+1)*size + col+3 + size*size)%(size*size);
    NLIST[4] = ((row-1)*size + col+2 + size*size)%(size*size);
    NLIST[5] = ((row-2)*size + col+1 + size*size)%(size*size);
    NLIST[6] = ((row+2)*size + col-1 + size*size)%(size*size);
    NLIST[7] = ((row+1)*size + col-2 + size*size)%(size*size);
    NLIST[8] = ((row-1)*size + col-3 + size*size)%(size*size);
    NLIST[9] = ((row+3)*size + col+2 + size*size)%(size*size);
    NLIST[10] = ((row+1)*size+ col+3 + size*size)%(size*size);
    NLIST[11] = ((row+2)*size+ col+3 + size*size)%(size*size);
}

__device__ void N1_4_4_4_8(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row-1)*size + col   + size*size)%(size*size);
    NLIST[1] = ((row  )*size + col-1 + size*size)%(size*size);
    NLIST[2] = ((row+1)*size + col   + size*size)%(size*size);
    NLIST[3] = ((row  )*size + col+1 + size*size)%(size*size);
}

__device__ void N2_4_4_4_8(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row-1)*size + col-1 + size*size)%(size*size);
    NLIST[1] = ((row-1)*size + col+1 + size*size)%(size*size);
    NLIST[2] = ((row+1)*size + col-1 + size*size)%(size*size);
    NLIST[3] = ((row+1)*size + col+1 + size*size)%(size*size);
}

__device__ void N3_4_4_4_8(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row-2)*size + col   + size*size)%(size*size);
    NLIST[1] = ((row  )*size + col-2 + size*size)%(size*size);
    NLIST[2] = ((row+2)*size + col   + size*size)%(size*size);
    NLIST[3] = ((row  )*size + col+2 + size*size)%(size*size);
}

__device__ void N4_4_4_4_8(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row-2)*size + col-1 + size*size)%(size*size);
    NLIST[1] = ((row-1)*size + col-2 + size*size)%(size*size);
    NLIST[2] = ((row-2)*size + col+1 + size*size)%(size*size);
    NLIST[3] = ((row-1)*size + col+2 + size*size)%(size*size);
    NLIST[4] = ((row+1)*size + col-2 + size*size)%(size*size);
    NLIST[5] = ((row+2)*size + col-1 + size*size)%(size*size);
    NLIST[6] = ((row+1)*size + col+2 + size*size)%(size*size);
    NLIST[7] = ((row+2)*size + col+1 + size*size)%(size*size);
}

__device__ void N1_2_2_4_2(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row-1)*size + col + size*size)%(size*size);
    NLIST[1] = ((row+1)*size + col + size*size)%(size*size);
}

__device__ void N2_2_2_4_2(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row+1)*size + col-1   + size*size)%(size*size);
    NLIST[1] = ((row+1)*size + col+1   + size*size)%(size*size);
}

__device__ void N3_2_2_4_2(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row+1)*size + (col-1+size)%size + size*size)%(size*size);
    NLIST[1] = ((row+1)*size + col+1 + size*size)%(size*size);
    NLIST[2] = (size - (row-1)*size + col-1 + size*size)%(size*size);
    NLIST[3] = ((row-1)*size + col+1 + size*size)%(size*size);
}

__device__ void N4_2_2_4_2(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row-2)*size + col   + size*size)%(size*size);
    NLIST[1] = ((row-2)*size + col   + size*size)%(size*size);
}

__device__ void N1_2_4_2_4(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = (row*size + col-1   + size*size)%(size*size);
    NLIST[1] = (row*size + col+1   + size*size)%(size*size);
}

__device__ void N2_2_4_2_4(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row-1)*size + col   + size*size)%(size*size);
    NLIST[1] = ((row+1)*size + col   + size*size)%(size*size);
    NLIST[2] = ((row-1)*size + col-1 + size*size)%(size*size);
    NLIST[3] = ((row+1)*size + col-1 + size*size)%(size*size);
}

__device__ void N3_2_4_2_4(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row-2)*size + col   + size*size)%(size*size);
    NLIST[1] = ((row+2)*size + col   + size*size)%(size*size);
}

__device__ void N4_2_4_2_4(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row-2)*size + col-1   + size*size)%(size*size);
    NLIST[1] = ((row+2)*size + col+1   + size*size)%(size*size);
    NLIST[2] = ((row-2)*size + col-1   + size*size)%(size*size);
    NLIST[3] = ((row+2)*size + col+1   + size*size)%(size*size);
}

__device__ void N1_3_6_3_6(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row+1)*size + col+0 + size*size)%(size*size);
    NLIST[1] = ((row-1)*size + col+1 + size*size)%(size*size);
    NLIST[2] = ((row-0)*size + col-1 + size*size)%(size*size);
}

__device__ void N2_3_6_3_6(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row+1)*size + col+1 + size*size)%(size*size);
    NLIST[1] = ((row-1)*size + col+2 + size*size)%(size*size);
    NLIST[2] = ((row-2)*size + col+1 + size*size)%(size*size);
    NLIST[3] = ((row-1)*size + col-1 + size*size)%(size*size);
    NLIST[4] = ((row+1)*size + col-2 + size*size)%(size*size);
    NLIST[5] = ((row+2)*size + col-1 + size*size)%(size*size);
}

__device__ void N3_3_6_3_6(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row+2)*size + col   + size*size)%(size*size);
    NLIST[1] = ((row-2)*size + col+2 + size*size)%(size*size);
    NLIST[2] = ((row  )*size + col-2 + size*size)%(size*size);
}

__device__ void N4_3_6_3_6(int n, int size, int* NLIST)
{
    int row = n/size;
    int col = n%size;
    NLIST[0] = ((row-1)*size + col+3 + size*size)%(size*size);
    NLIST[1] = ((row-3)*size + col+1 + size*size)%(size*size);
    NLIST[2] = ((row-2)*size + col-1 + size*size)%(size*size);
    NLIST[3] = ((row+2)*size + col-3 + size*size)%(size*size);
    NLIST[4] = ((row-3)*size + col+2 + size*size)%(size*size);
    NLIST[5] = ((row+1)*size + col+2 + size*size)%(size*size);
}

//Hamiltonians
__device__ float_t hamiltonian_tc_2d_6_6_6_12_dm1(float_t* mat, float_t* sheet, int pti, float_t spinx, float_t spiny, float_t spinz, float_t* NVEC, float_t* b, int size)
{
    float_t H = 0.0;
    float_t d_plane = mat[25];

    int n1list[6];
    int n2list[6];
    int n3list[6];
    int n4list[12];

    N1_6_6_6_12(pti, size, n1list);
    N2_6_6_6_12(pti, size, n2list);
    N3_6_6_6_12(pti, size, n3list);
    N4_6_6_6_12(pti, size, n4list);

    for(int i=0; i<6; i++)
    {
        H += mat[1]*(spinx*sheet[n1list[i]*3] + spiny*sheet[n1list[i]*3+1] + spinz*sheet[n1list[i]*3+2]) + mat[5]*spinx*sheet[n1list[i]*3] + mat[6]*spiny*sheet[n1list[i]*3+1] + mat[7]*spinz*sheet[n1list[i]*3+2];
        float_t CRijx = (spinz*sheet[n1list[i]*3+1] - spiny*sheet[n1list[i]*3+2]) * NVEC[i*3];
        float_t CRijy = (spinx*sheet[n1list[i]*3+2] - spinz*sheet[n1list[i]*3]) * NVEC[i*3+1];
        float_t CRijz = (spiny*sheet[n1list[i]*3] - spinx*sheet[n1list[i]*3+1]) * NVEC[i*3+2];
        H += d_plane*(CRijx+CRijy+CRijz);
    }
    for(int i=0; i<6; i++)
    {
        H += mat[2]*(spinx*sheet[n2list[i]*3] + spiny*sheet[n2list[i]*3+1] + spinz*sheet[n2list[i]*3+2]) + mat[8]*spinx*sheet[n2list[i]*3] + mat[9]*spiny*sheet[n2list[i]*3+1] + mat[10]*spinz*sheet[n2list[i]*3+2];
    }
    for(int i=0; i<6; i++)
    {
        H += mat[3]*(spinx*sheet[n3list[i]*3] + spiny*sheet[n3list[i]*3+1] + spinz*sheet[n3list[i]*3+2]) + mat[11]*spinx*sheet[n3list[i]*3] + mat[12]*spiny*sheet[n3list[i]*3+1] + mat[13]*spinz*sheet[n3list[i]*3+2];
    }
    for(int i=0; i<12; i++)
    {
        H += mat[4]*(spinx*sheet[n4list[i]*3] + spiny*sheet[n4list[i]*3+1] + spinz*sheet[n4list[i]*3+2]) + mat[14]*spinx*sheet[n4list[i]*3] + mat[15]*spiny*sheet[n4list[i]*3+1] + mat[16]*spinz*sheet[n4list[i]*3+2];
    }
    H += mat[17]*spinx*spinx + mat[18]*spiny*spiny + mat[19]*spinz*spinz;
    H += b[0]*spinz;
    return -1.0*H;
}

__device__ float_t hamiltonian_tc_2d_6_6_6_12_dm0(float_t* mat, float_t* sheet, int pti, float_t spinx, float_t spiny, float_t spinz, float_t* b, int size)
{
    float_t H = 0.0;

    int n1list[6];
    int n2list[6];
    int n3list[6];
    int n4list[12];
    
    N1_6_6_6_12(pti, size, n1list);
    N2_6_6_6_12(pti, size, n2list);
    N3_6_6_6_12(pti, size, n3list);
    N4_6_6_6_12(pti, size, n4list);

    for(int i=0; i<6; i++)
    {
        H += mat[1]*(spinx*sheet[n1list[i]*3] + spiny*sheet[n1list[i]*3+1] + spinz*sheet[n1list[i]*3+2]) + mat[5]*spinx*sheet[n1list[i]*3] + mat[6]*spiny*sheet[n1list[i]*3+1] + mat[7]*spinz*sheet[n1list[i]*3+2];
    }
    for(int i=0; i<6; i++)
    {
        H += mat[2]*(spinx*sheet[n2list[i]*3] + spiny*sheet[n2list[i]*3+1] + spinz*sheet[n2list[i]*3+2]) + mat[8]*spinx*sheet[n2list[i]*3] + mat[9]*spiny*sheet[n2list[i]*3+1] + mat[10]*spinz*sheet[n2list[i]*3+2];
    }
    for(int i=0; i<6; i++)
    {
        H += mat[3]*(spinx*sheet[n3list[i]*3] + spiny*sheet[n3list[i]*3+1] + spinz*sheet[n3list[i]*3+2]) + mat[11]*spinx*sheet[n3list[i]*3] + mat[12]*spiny*sheet[n3list[i]*3+1] + mat[13]*spinz*sheet[n3list[i]*3+2];
    }
    for(int i=0; i<12; i++)
    {
        H += mat[4]*(spinx*sheet[n4list[i]*3] + spiny*sheet[n4list[i]*3+1] + spinz*sheet[n4list[i]*3+2]) + mat[14]*spinx*sheet[n4list[i]*3] + mat[15]*spiny*sheet[n4list[i]*3+1] + mat[16]*spinz*sheet[n4list[i]*3+2];
    }
    H += mat[17]*spinx*spinx + mat[18]*spiny*spiny + mat[19]*spinz*spinz;
    H += b[0]*spinz;
    return -1.0*H;
}

__device__ float_t hamiltonian_tc_2d_4_4_4_8_dm1(float_t* mat, float_t* sheet, int pti, float_t spinx, float_t spiny, float_t spinz, float_t* NVEC, float_t* b, int size)
{
    float_t H = 0.0;
    float_t d_plane = mat[25];

    int n1list[6];
    int n2list[6];
    int n3list[6];
    int n4list[12];

    N1_4_4_4_8(pti, size, n1list);
    N2_4_4_4_8(pti, size, n2list);
    N3_4_4_4_8(pti, size, n3list);
    N4_4_4_4_8(pti, size, n4list);

    for(int i=0; i<6; i++)
    {
        H += mat[1]*(spinx*sheet[n1list[i]*3] + spiny*sheet[n1list[i]*3+1] + spinz*sheet[n1list[i]*3+2]) + mat[5]*spinx*sheet[n1list[i]*3] + mat[6]*spiny*sheet[n1list[i]*3+1] + mat[7]*spinz*sheet[n1list[i]*3+2];
        float_t CRijx = (spinz*sheet[n1list[i]*3+1] - spiny*sheet[n1list[i]*3+2]) * NVEC[i*3];
        float_t CRijy = (spinx*sheet[n1list[i]*3+2] - spinz*sheet[n1list[i]*3]) * NVEC[i*3+1];
        float_t CRijz = (spiny*sheet[n1list[i]*3] - spinx*sheet[n1list[i]*3+1]) * NVEC[i*3+2];
        H += d_plane*(CRijx+CRijy+CRijz);
    }
    for(int i=0; i<6; i++)
    {
        H += mat[2]*(spinx*sheet[n2list[i]*3] + spiny*sheet[n2list[i]*3+1] + spinz*sheet[n2list[i]*3+2]) + mat[8]*spinx*sheet[n2list[i]*3] + mat[9]*spiny*sheet[n2list[i]*3+1] + mat[10]*spinz*sheet[n2list[i]*3+2];
    }
    for(int i=0; i<6; i++)
    {
        H += mat[3]*(spinx*sheet[n3list[i]*3] + spiny*sheet[n3list[i]*3+1] + spinz*sheet[n3list[i]*3+2]) + mat[11]*spinx*sheet[n3list[i]*3] + mat[12]*spiny*sheet[n3list[i]*3+1] + mat[13]*spinz*sheet[n3list[i]*3+2];
    }
    for(int i=0; i<12; i++)
    {
        H += mat[4]*(spinx*sheet[n4list[i]*3] + spiny*sheet[n4list[i]*3+1] + spinz*sheet[n4list[i]*3+2]) + mat[14]*spinx*sheet[n4list[i]*3] + mat[15]*spiny*sheet[n4list[i]*3+1] + mat[16]*spinz*sheet[n4list[i]*3+2];
    }
    H += mat[17]*spinx*spinx + mat[18]*spiny*spiny + mat[19]*spinz*spinz;
    H += b[0]*spinz;
    return -1.0*H;
}

__device__ float_t hamiltonian_tc_2d_4_4_4_8_dm0(float_t* mat, float_t* sheet, int pti, float_t spinx, float_t spiny, float_t spinz, float_t* b, int size)
{
    float_t H = 0.0;

    int n1list[6];
    int n2list[6];
    int n3list[6];
    int n4list[12];
    
    N1_4_4_4_8(pti, size, n1list);
    N2_4_4_4_8(pti, size, n2list);
    N3_4_4_4_8(pti, size, n3list);
    N4_4_4_4_8(pti, size, n4list);

    for(int i=0; i<6; i++)
    {
        H += mat[1]*(spinx*sheet[n1list[i]*3] + spiny*sheet[n1list[i]*3+1] + spinz*sheet[n1list[i]*3+2]) + mat[5]*spinx*sheet[n1list[i]*3] + mat[6]*spiny*sheet[n1list[i]*3+1] + mat[7]*spinz*sheet[n1list[i]*3+2];
    }
    for(int i=0; i<6; i++)
    {
        H += mat[2]*(spinx*sheet[n2list[i]*3] + spiny*sheet[n2list[i]*3+1] + spinz*sheet[n2list[i]*3+2]) + mat[8]*spinx*sheet[n2list[i]*3] + mat[9]*spiny*sheet[n2list[i]*3+1] + mat[10]*spinz*sheet[n2list[i]*3+2];
    }
    for(int i=0; i<6; i++)
    {
        H += mat[3]*(spinx*sheet[n3list[i]*3] + spiny*sheet[n3list[i]*3+1] + spinz*sheet[n3list[i]*3+2]) + mat[11]*spinx*sheet[n3list[i]*3] + mat[12]*spiny*sheet[n3list[i]*3+1] + mat[13]*spinz*sheet[n3list[i]*3+2];
    }
    for(int i=0; i<12; i++)
    {
        H += mat[4]*(spinx*sheet[n4list[i]*3] + spiny*sheet[n4list[i]*3+1] + spinz*sheet[n4list[i]*3+2]) + mat[14]*spinx*sheet[n4list[i]*3] + mat[15]*spiny*sheet[n4list[i]*3+1] + mat[16]*spinz*sheet[n4list[i]*3+2];
    }
    H += mat[17]*spinx*spinx + mat[18]*spiny*spiny + mat[19]*spinz*spinz;
    H += b[0]*spinz;
    return -1.0*H;
}

__device__ float_t hamiltonian_tc_2d_3_6_3_6_dm0(float_t* mat, float_t* sheet, int pti, float_t spinx, float_t spiny, float_t spinz, float_t* b, int size)
{
    float_t H = 0.0;

    int n1list[3];
    int n2list[6];
    int n3list[3];
    int n4list[6];

    N1_3_6_3_6(pti, size, n1list);
    N2_3_6_3_6(pti, size, n2list);
    N3_3_6_3_6(pti, size, n3list);
    N4_3_6_3_6(pti, size, n4list);

    for(int i=0; i<3; i++)
    {
        H += mat[1]*(spinx*sheet[n1list[i]*3] + spiny*sheet[n1list[i]*3+1] + spinz*sheet[n1list[i]*3+2]) + mat[5]*spinx*sheet[n1list[i]*3] + mat[6]*spiny*sheet[n1list[i]*3+1] + mat[7]*spinz*sheet[n1list[i]*3+2];
    }
    for(int i=0; i<6; i++)
    {
        H += mat[2]*(spinx*sheet[n2list[i]*3] + spiny*sheet[n2list[i]*3+1] + spinz*sheet[n2list[i]*3+2]) + mat[8]*spinx*sheet[n2list[i]*3] + mat[9]*spiny*sheet[n2list[i]*3+1] + mat[10]*spinz*sheet[n2list[i]*3+2];
    }
    for(int i=0; i<3; i++)
    {
        H += mat[3]*(spinx*sheet[n3list[i]*3] + spiny*sheet[n3list[i]*3+1] + spinz*sheet[n3list[i]*3+2]) + mat[11]*spinx*sheet[n3list[i]*3] + mat[12]*spiny*sheet[n3list[i]*3+1] + mat[13]*spinz*sheet[n3list[i]*3+2];
    }
    for(int i=0; i<6; i++)
    {
        H += mat[4]*(spinx*sheet[n4list[i]*3] + spiny*sheet[n4list[i]*3+1] + spinz*sheet[n4list[i]*3+2]) + mat[14]*spinx*sheet[n4list[i]*3] + mat[15]*spiny*sheet[n4list[i]*3+1] + mat[16]*spinz*sheet[n4list[i]*3+2];
    }
    H += mat[17]*spinx*spinx + mat[18]*spiny*spiny + mat[19]*spinz*spinz;
    H += b[0]*spinz;
    return -1.0*H;
}

__device__ float_t hamiltonian_tc_2d_3_6_3_6_dm1(float_t* mat, float_t* sheet, int pti, float_t spinx, float_t spiny, float_t spinz, float_t* NVEC, float_t* b, int size)
{
    float_t H = 0.0;
    float_t d_plane = mat[25];

    int n1list[3];
    int n2list[6];
    int n3list[3];
    int n4list[6];

    N1_3_6_3_6(pti, size, n1list);
    N2_3_6_3_6(pti, size, n2list);
    N3_3_6_3_6(pti, size, n3list);
    N4_3_6_3_6(pti, size, n4list);

    for(int i=0; i<3; i++)
    {
        H += mat[1]*(spinx*sheet[n1list[i]*3] + spiny*sheet[n1list[i]*3+1] + spinz*sheet[n1list[i]*3+2]) + mat[5]*spinx*sheet[n1list[i]*3] + mat[6]*spiny*sheet[n1list[i]*3+1] + mat[7]*spinz*sheet[n1list[i]*3+2];
        float_t CRijx = (spinz*sheet[n1list[i]*3+1] - spiny*sheet[n1list[i]*3+2]) * NVEC[i*3];
        float_t CRijy = (spinx*sheet[n1list[i]*3+2] - spinz*sheet[n1list[i]*3]) * NVEC[i*3+1];
        float_t CRijz = (spiny*sheet[n1list[i]*3] - spinx*sheet[n1list[i]*3+1]) * NVEC[i*3+2];
        H += d_plane*(CRijx+CRijy+CRijz);
    }
    for(int i=0; i<6; i++)
    {
        H += mat[2]*(spinx*sheet[n2list[i]*3] + spiny*sheet[n2list[i]*3+1] + spinz*sheet[n2list[i]*3+2]) + mat[8]*spinx*sheet[n2list[i]*3] + mat[9]*spiny*sheet[n2list[i]*3+1] + mat[10]*spinz*sheet[n2list[i]*3+2];
    }
    for(int i=0; i<3; i++)
    {
        H += mat[3]*(spinx*sheet[n3list[i]*3] + spiny*sheet[n3list[i]*3+1] + spinz*sheet[n3list[i]*3+2]) + mat[11]*spinx*sheet[n3list[i]*3] + mat[12]*spiny*sheet[n3list[i]*3+1] + mat[13]*spinz*sheet[n3list[i]*3+2];
    }
    for(int i=0; i<6; i++)
    {
        H += mat[4]*(spinx*sheet[n4list[i]*3] + spiny*sheet[n4list[i]*3+1] + spinz*sheet[n4list[i]*3+2]) + mat[14]*spinx*sheet[n4list[i]*3] + mat[15]*spiny*sheet[n4list[i]*3+1] + mat[16]*spinz*sheet[n4list[i]*3+2];
    }
    H += mat[17]*spinx*spinx + mat[18]*spiny*spiny + mat[19]*spinz*spinz;
    H += b[0]*spinz;
    return -1.0*H;
}

__device__ float_t hamiltonian_tc_2d_3_6_3_6_dm2(float_t* mat, float_t* sheet, int pti, float_t spinx, float_t spiny, float_t spinz, float_t* b, int size)
{
    float_t H = 0.0;

    int n1list[3];
    int n2list[6];
    int n3list[3];
    int n4list[6];

    N1_3_6_3_6(pti, size, n1list);
    N2_3_6_3_6(pti, size, n2list);
    N3_3_6_3_6(pti, size, n3list);
    N4_3_6_3_6(pti, size, n4list);

    for(int i=0; i<3; i++)
    {
        H += mat[1]*(spinx*sheet[n1list[i]*3] + spiny*sheet[n1list[i]*3+1] + spinz*sheet[n1list[i]*3+2]) + mat[5]*spinx*sheet[n1list[i]*3] + mat[6]*spiny*sheet[n1list[i]*3+1] + mat[7]*spinz*sheet[n1list[i]*3+2];
        H += 0.21*(spinx*sheet[n1list[i]*3] + spiny*sheet[n1list[i]*3+1] + spinz*sheet[n1list[i]*3+2])*(spinx*sheet[n1list[i]*3] + spiny*sheet[n1list[i]*3+1] + spinz*sheet[n1list[i]*3+2]);
    }
    for(int i=0; i<6; i++)
    {
        H += mat[2]*(spinx*sheet[n2list[i]*3] + spiny*sheet[n2list[i]*3+1] + spinz*sheet[n2list[i]*3+2]) + mat[8]*spinx*sheet[n2list[i]*3] + mat[9]*spiny*sheet[n2list[i]*3+1] + mat[10]*spinz*sheet[n2list[i]*3+2];
    }
    for(int i=0; i<3; i++)
    {
        H += mat[3]*(spinx*sheet[n3list[i]*3] + spiny*sheet[n3list[i]*3+1] + spinz*sheet[n3list[i]*3+2]) + mat[11]*spinx*sheet[n3list[i]*3] + mat[12]*spiny*sheet[n3list[i]*3+1] + mat[13]*spinz*sheet[n3list[i]*3+2];
    }
    for(int i=0; i<6; i++)
    {
        H += mat[4]*(spinx*sheet[n4list[i]*3] + spiny*sheet[n4list[i]*3+1] + spinz*sheet[n4list[i]*3+2]) + mat[14]*spinx*sheet[n4list[i]*3] + mat[15]*spiny*sheet[n4list[i]*3+1] + mat[16]*spinz*sheet[n4list[i]*3+2];
    }
    H += mat[17]*spinx*spinx + mat[18]*spiny*spiny + mat[19]*spinz*spinz;
    H += b[0]*spinz;
    return -1.0*H;
}


__device__ float_t hamiltonian_tc_2d_2_2_4_2_dm0(float_t* mat, float_t* sheet, int pti, float_t spinx, float_t spiny, float_t spinz, float_t* b, int size)
{
    float_t H = 0.0;

    int n1list[2];
    int n2list[2];
    int n3list[4];
    int n4list[2];

    N1_2_2_4_2(pti, size, n1list);
    N2_2_2_4_2(pti, size, n2list);
    N3_2_2_4_2(pti, size, n3list);
    N4_2_2_4_2(pti, size, n4list);

    for(int i=0; i<2; i++)
    {
        H += mat[1]*(spinx*sheet[n1list[i]*3] + spiny*sheet[n1list[i]*3+1] + spinz*sheet[n1list[i]*3+2]) + mat[5]*spinx*sheet[n1list[i]*3] + mat[6]*spiny*sheet[n1list[i]*3+1] + mat[7]*spinz*sheet[n1list[i]*3+2];
    }
    for(int i=0; i<2; i++)
    {
        H += mat[2]*(spinx*sheet[n2list[i]*3] + spiny*sheet[n2list[i]*3+1] + spinz*sheet[n2list[i]*3+2]) + mat[8]*spinx*sheet[n2list[i]*3] + mat[9]*spiny*sheet[n2list[i]*3+1] + mat[10]*spinz*sheet[n2list[i]*3+2];
    }
    for(int i=0; i<4; i++)
    {
        H += mat[3]*(spinx*sheet[n3list[i]*3] + spiny*sheet[n3list[i]*3+1] + spinz*sheet[n3list[i]*3+2]) + mat[11]*spinx*sheet[n3list[i]*3] + mat[12]*spiny*sheet[n3list[i]*3+1] + mat[13]*spinz*sheet[n3list[i]*3+2];
    }
    for(int i=0; i<2; i++)
    {
        H += mat[4]*(spinx*sheet[n4list[i]*3] + spiny*sheet[n4list[i]*3+1] + spinz*sheet[n4list[i]*3+2]) + mat[14]*spinx*sheet[n4list[i]*3] + mat[15]*spiny*sheet[n4list[i]*3+1] + mat[16]*spinz*sheet[n4list[i]*3+2];
    }
    H += mat[17]*spinx*spinx + mat[18]*spiny*spiny + mat[19]*spinz*spinz;
    H += b[0]*spinz;
    return -1.0*H;
}

__device__ float_t hamiltonian_tc_2d_2_4_2_4_dm0(float_t* mat, float_t* sheet, int pti, float_t spinx, float_t spiny, float_t spinz, float_t* b, int size)
{
    float_t H = 0.0;

    int n1list[2];
    int n2list[4];
    int n3list[2];
    int n4list[4];

    N1_2_4_2_4(pti, size, n1list);
    N2_2_4_2_4(pti, size, n2list);
    N3_2_4_2_4(pti, size, n3list);
    N4_2_4_2_4(pti, size, n4list);

    for(int i=0; i<2; i++)
    {
        H += mat[1]*(spinx*sheet[n1list[i]*3] + spiny*sheet[n1list[i]*3+1] + spinz*sheet[n1list[i]*3+2]) + mat[5]*spinx*sheet[n1list[i]*3] + mat[6]*spiny*sheet[n1list[i]*3+1] + mat[7]*spinz*sheet[n1list[i]*3+2];
    }
    for(int i=0; i<4; i++)
    {
        H += mat[2]*(spinx*sheet[n2list[i]*3] + spiny*sheet[n2list[i]*3+1] + spinz*sheet[n2list[i]*3+2]) + mat[8]*spinx*sheet[n2list[i]*3] + mat[9]*spiny*sheet[n2list[i]*3+1] + mat[10]*spinz*sheet[n2list[i]*3+2];
    }
    for(int i=0; i<2; i++)
    {
        H += mat[3]*(spinx*sheet[n3list[i]*3] + spiny*sheet[n3list[i]*3+1] + spinz*sheet[n3list[i]*3+2]) + mat[11]*spinx*sheet[n3list[i]*3] + mat[12]*spiny*sheet[n3list[i]*3+1] + mat[13]*spinz*sheet[n3list[i]*3+2];
    }
    for(int i=0; i<4; i++)
    {
        H += mat[4]*(spinx*sheet[n4list[i]*3] + spiny*sheet[n4list[i]*3+1] + spinz*sheet[n4list[i]*3+2]) + mat[14]*spinx*sheet[n4list[i]*3] + mat[15]*spiny*sheet[n4list[i]*3+1] + mat[16]*spinz*sheet[n4list[i]*3+2];
    }
    H += mat[17]*spinx*spinx + mat[18]*spiny*spiny + mat[19]*spinz*spinz;
    H += b[0]*spinz;
    return -1.0*H;
}
//Monte Carlo
__global__ void metropolis_mc_dm1_6_6_6_12(float_t *mat, float_t *sheet, float_t *T, int* N, float_t* S1, float_t* S2,float_t* S3, float_t* R, float_t* tf, float_t* NVEC, float_t* B, int* size)
{
    __shared__ float_t L0;
    __shared__ float_t L1;
    __shared__ float_t dE;
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = N[threadID];
    int tidx = threadIdx.x;
    if (tidx == 0)
    {  
        L0 = hamiltonian_tc_2d_6_6_6_12_dm1(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2], NVEC, B, size[0]);
    }
    if (tidx == 1)
    {
        L1 = hamiltonian_tc_2d_6_6_6_12_dm1(mat, sheet, pt_thread, S1[threadID], S2[threadID], S3[threadID], NVEC, B, size[0]);
    }
    __syncthreads();

    dE = L1 - L0;
    
    if (dE < 0)
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
    else if (expf(-1.0*dE*T[0]) > R[threadID])
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
}

__global__ void metropolis_mc_dm0_6_6_6_12(float_t *mat, float_t *sheet, float_t *T, int* N, float_t* S1, float_t* S2,float_t* S3, float_t* R, float_t* tf, float_t* B, int* size)
{
    __shared__ float_t L0;
    __shared__ float_t L1;
    __shared__ float_t dE;
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = N[threadID];
    int tidx = threadIdx.x;
    if (tidx == 0)
    {  
        L0 = hamiltonian_tc_2d_6_6_6_12_dm0(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2],  B, size[0]);
    }
    if (tidx == 1)
    {
        L1 = hamiltonian_tc_2d_6_6_6_12_dm0(mat, sheet, pt_thread, S1[threadID], S2[threadID], S3[threadID], B, size[0]);
    }
    __syncthreads();

    dE = L1 - L0;
    
    if (dE < 0)
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
    else if (expf(-1.0*dE*T[0]) > R[threadID])
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
}

__global__ void metropolis_mc_dm0_3_6_3_6(float_t *mat, float_t *sheet, float_t *T, int* N, float_t* S1, float_t* S2,float_t* S3, float_t* R, float_t* tf, float_t* B, int* size)
{
    __shared__ float_t L0;
    __shared__ float_t L1;
    __shared__ float_t dE;
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = N[threadID];
    int tidx = threadIdx.x;
    if (tidx == 0)
    {  
        L0 = hamiltonian_tc_2d_3_6_3_6_dm0(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2],  B, size[0]);
    }
    if (tidx == 1)
    {
        L1 = hamiltonian_tc_2d_3_6_3_6_dm0(mat, sheet, pt_thread, S1[threadID], S2[threadID], S3[threadID], B, size[0]);
    }
    __syncthreads();

    dE = L1 - L0;

    if (dE < 0)
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
    else if (expf(-1.0*dE*T[0]) > R[threadID])
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
}

__global__ void metropolis_mc_dm1_3_6_3_6(float_t *mat, float_t *sheet, float_t *T, int* N, float_t* S1, float_t* S2,float_t* S3, float_t* R, float_t* tf, float_t* NVEC, float_t* B, int* size)
{
    __shared__ float_t L0;
    __shared__ float_t L1;
    __shared__ float_t dE;
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = N[threadID];
    int tidx = threadIdx.x;
    if (tidx == 0)
    {  
        L0 = hamiltonian_tc_2d_3_6_3_6_dm1(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2], NVEC, B, size[0]);
    }
    if (tidx == 1)
    {
        L1 = hamiltonian_tc_2d_3_6_3_6_dm1(mat, sheet, pt_thread, S1[threadID], S2[threadID], S3[threadID], NVEC, B, size[0]);
    }
    __syncthreads();

    dE = L1 - L0;

    if (dE < 0)
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
    else if (expf(-1.0*dE*T[0]) > R[threadID])
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
}

__global__ void metropolis_mc_dm2_3_6_3_6(float_t *mat, float_t *sheet, float_t *T, int* N, float_t* S1, float_t* S2,float_t* S3, float_t* R, float_t* tf, float_t* B, int* size)
{
    __shared__ float_t L0;
    __shared__ float_t L1;
    __shared__ float_t dE;
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = N[threadID];
    int tidx = threadIdx.x;
    if (tidx == 0)
    {  
        L0 = hamiltonian_tc_2d_3_6_3_6_dm2(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2],  B, size[0]);
    }
    if (tidx == 1)
    {
        L1 = hamiltonian_tc_2d_3_6_3_6_dm2(mat, sheet, pt_thread, S1[threadID], S2[threadID], S3[threadID], B, size[0]);
    }
    __syncthreads();

    dE = L1 - L0;

    if (dE < 0)
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
    else if (expf(-1.0*dE*T[0]) > R[threadID])
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
}

__global__ void metropolis_mc_dm1_4_4_4_8(float_t *mat, float_t *sheet, float_t *T, int* N, float_t* S1, float_t* S2,float_t* S3, float_t* R, float_t* tf, float_t* NVEC, float_t* B, int* size)
{
    __shared__ float_t L0;
    __shared__ float_t L1;
    __shared__ float_t dE;
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = N[threadID];
    int tidx = threadIdx.x;
    if (tidx == 0)
    {  
        L0 = hamiltonian_tc_2d_4_4_4_8_dm1(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2], NVEC, B, size[0]);
    }
    if (tidx == 1)
    {
        L1 = hamiltonian_tc_2d_4_4_4_8_dm1(mat, sheet, pt_thread, S1[threadID], S2[threadID], S3[threadID], NVEC, B, size[0]);
    }
    __syncthreads();

    dE = L1 - L0;
    
    if (dE < 0)
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
    else if (expf(-1.0*dE*T[0]) > R[threadID])
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
}

__global__ void metropolis_mc_dm0_4_4_4_8(float_t *mat, float_t *sheet, float_t *T, int* N, float_t* S1, float_t* S2,float_t* S3, float_t* R, float_t* tf, float_t* NVEC, float_t* B, int* size)
{
    __shared__ float_t L0;
    __shared__ float_t L1;
    __shared__ float_t dE;
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = N[threadID];
    int tidx = threadIdx.x;
    if (tidx == 0)
    {  
        L0 = hamiltonian_tc_2d_4_4_4_8_dm0(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2], B, size[0]);
    }
    if (tidx == 1)
    {
        L1 = hamiltonian_tc_2d_4_4_4_8_dm0(mat, sheet, pt_thread, S1[threadID], S2[threadID], S3[threadID], B, size[0]);
    }
    __syncthreads();

    dE = L1 - L0;
    
    if (dE < 0)
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
    else if (expf(-1.0*dE*T[0]) > R[threadID])
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
}

__global__ void metropolis_mc_dm0_2_2_4_2(float_t *mat, float_t *sheet, float_t *T, int* N, float_t* S1, float_t* S2, float_t* S3, float_t* R, float_t* tf, float_t* B, int* size)
{
    __shared__ float_t L0;
    __shared__ float_t L1;
    __shared__ float_t dE;
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = N[threadID];
    int tidx = threadIdx.x;
    if (tidx == 0)
    {
        L0 = hamiltonian_tc_2d_2_2_4_2_dm0(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2], B, size[0]);
    }
    if (tidx == 1)
    {
        L1 = hamiltonian_tc_2d_2_2_4_2_dm0(mat, sheet, pt_thread, S1[threadID], S2[threadID], S3[threadID], B, size[0]);
    }

    __syncthreads();

    dE = L1 - L0;

    if (dE < 0)
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
    else if (expf(-1.0*dE*T[0]) > R[threadID])
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
}

__global__ void metropolis_mc_dm0_2_4_2_4(float_t *mat, float_t *sheet, float_t *T, int* N, float_t* S1, float_t* S2, float_t* S3, float_t* R, float_t* tf, float_t* B, int* size)
{
    __shared__ float_t L0;
    __shared__ float_t L1;
    __shared__ float_t dE;

    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = N[threadID];
    int tidx = threadIdx.x;

    if (tidx == 0)
    {
        L0 = hamiltonian_tc_2d_2_4_2_4_dm0(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2], B, size[0]);
    }
    if (tidx == 1)
    {
        L1 = hamiltonian_tc_2d_2_4_2_4_dm0(mat, sheet, pt_thread, S1[threadID], S2[threadID], S3[threadID], B, size[0]);
    }

    __syncthreads();

    dE = L1 - L0;

    if (dE < 0)
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
    else if (expf(-1.0*dE*T[0]) > R[threadID])
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
    }
}
//Double Material Study
__device__ int alt_populate(float_t* sheet, int pt, int size)
{
    int n1list[3];
    N1_3_6_3_6(pt, size, n1list);
    if((sheet[pt*4+3] - 1)*(sheet[pt*4+3] - 1) <= 0.0001)
    {
        for(int j = 0; j < 3; j++)
        {
            if((sheet[n1list[j]*4+3] - 2)*(sheet[n1list[j]*4+3] - 2) <= 0.0001)
            {
                return 1;
            }
            else    
            {    
                sheet[n1list[j]*4+3] = 2;
                int s = alt_populate(sheet, n1list[j], size);
                return s;
            }
        }
    } 
    else if((sheet[pt*4+3] - 2)*(sheet[pt*4+3] - 2) <= 0.0001)
    {
        for(int j = 0; j < 3; j++)
        {
            if((sheet[n1list[j]*4+3] - 1)*(sheet[n1list[j]*4+3] - 1) <= 0.0001)
            {
                return 1;
            }
            else    
            {
                sheet[n1list[j]*4+3] = 1;
                int s = alt_populate(sheet, n1list[j], size);
                return s;
            }
        }
    }
}

__global__ void alt_grid(int* size, float_t* sheet, int* debug, float_t* spins)
{
    sheet[3] = 1;
    debug[0] = alt_populate(sheet, 0, size[0]);
    for(int i=0;i<size[0]*size[0];i++)
    {
        if(sheet[i*4+3]==1)
        {
            sheet[i*4] *= spins[0];
            sheet[i*4+1] *= spins[0];
            sheet[i*4+2] *= spins[0];
        }
        else if(sheet[i*4+3]==2)
        {
            sheet[i*4] *= spins[1];
            sheet[i*4+1] *= spins[1];
            sheet[i*4+2] *= spins[1];
        }
    }
}

//Energy Calculators
__global__ void energy_metropolis_mc_dm0_3_6_3_6(float_t *mat, float_t *sheet, float_t *T, int* N, float_t* S1, float_t* S2,float_t* S3, float_t* R, float_t* tf, float_t* B, int* size, float_t* en)
{
    __shared__ float_t L0;
    __shared__ float_t L1;
    __shared__ float_t dE;
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = N[threadID];
    int tidx = threadIdx.x;
    if (tidx == 0)
    {  
        L0 = hamiltonian_tc_2d_3_6_3_6_dm0(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2],  B, size[0]);
    }
    if (tidx == 1)
    {
        L1 = hamiltonian_tc_2d_3_6_3_6_dm0(mat, sheet, pt_thread, S1[threadID], S2[threadID], S3[threadID], B, size[0]);
    }
    __syncthreads();

    dE = L1 - L0;

    if (dE < 0)
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
        en[0] += dE;
    }
    else if (expf(-1.0*dE*T[0]) > R[threadID])
    {
        tf[threadID*4] = pt_thread;
        tf[threadID*4+1] = S1[threadID];
        tf[threadID*4+2] = S2[threadID];
        tf[threadID*4+3] = S3[threadID];
        en[0] += dE;
    }
}

__global__ void encalc_3636(float_t* mat, float_t* sheet, float_t* B, int* N, int* size, float_t* en)
{
    
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = threadID;
    en[idx] = -1.0*hamiltonian_tc_2d_3_6_3_6_dm0(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2], B, size[0]);
}

__global__ void encalc_3636_2(float_t* mat, float_t* sheet, float_t* B, int* N, int* size, float_t* en, float_t* NVEC)
{
    
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = threadID;
    en[idx] = -1.0*hamiltonian_tc_2d_3_6_3_6_dm1(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2], NVEC, B, size[0]);
}

__global__ void encalc_66612(float_t* mat, float_t* sheet, float_t* B, int* size, float_t* en)
{
    
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = threadID;
    en[idx] = -1.0*hamiltonian_tc_2d_6_6_6_12_dm0(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2], B, size[0]);
}

__global__ void encalc_66612_2(float_t* mat, float_t* sheet, float_t* B, int* size, float_t* en, float_t* NVEC)
{
    
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = threadID;
    en[idx] = -1.0*hamiltonian_tc_2d_6_6_6_12_dm1(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2], NVEC, B, size[0]);
}

__global__ void encalc_4448(float_t* mat, float_t* sheet, float_t* B, int* size, float_t* en)
{
    
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = threadID;
    en[idx] = -1.0*hamiltonian_tc_2d_4_4_4_8_dm0(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2], B, size[0]);
}

__global__ void encalc_4448_2(float_t* mat, float_t* sheet, float_t* B, int* size, float_t* en, float_t* NVEC)
{
    
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = threadID;
    en[idx] = -1.0*hamiltonian_tc_2d_4_4_4_8_dm1(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2], NVEC, B, size[0]);
}

__global__ void encalc_2424(float_t* mat, float_t* sheet, float_t* B, int* size, float_t* en)
{
    
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = threadID;
    en[idx] = -1.0*hamiltonian_tc_2d_2_4_2_4_dm0(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2], B, size[0]);
}

__global__ void encalc_2242(float_t* mat, float_t* sheet, float_t* B, int* size, float_t* en)
{
    
    int idx = blockIdx.x;
    int threadID = idx;
    int pt_thread = threadID;
    en[idx] = -1.0*hamiltonian_tc_2d_2_2_4_2_dm0(mat, sheet, pt_thread, sheet[pt_thread*3], sheet[pt_thread*3+1], sheet[pt_thread*3+2], B, size[0]);
}

//!cuda
""")
#KERNEL CODE END
# =============================================================================


# =============================================================================
# CUDA KERNEL FUNCTION DEFINITIONS
# In case of any additions to the CUDA kernel code, the functions need to be
# defined here.
# =============================================================================
#CUDA KERNEL FUNCTION DEFINITIONS
GRID_COPY = dev_hamiltonian.get_function("cp_grid")
ALT_GRID_COPY = dev_hamiltonian.get_function("alt_cp_grid")

NPREC = dev_hamiltonian.get_function("NList_processor")
VPREC = dev_hamiltonian.get_function("uvec_processor")
ISING = dev_hamiltonian.get_function("alt_uvec_ising")
ALT_GRID = dev_hamiltonian.get_function("alt_grid")

METROPOLIS_MC_DM0_2_2_4_2 = dev_hamiltonian.get_function("metropolis_mc_dm0_2_2_4_2")
METROPOLIS_MC_DM0_2_4_2_4 = dev_hamiltonian.get_function("metropolis_mc_dm0_2_4_2_4")

METROPOLIS_MC_DM0_3_6_3_6 = dev_hamiltonian.get_function("metropolis_mc_dm0_3_6_3_6")
METROPOLIS_MC_DM1_3_6_3_6 = dev_hamiltonian.get_function("metropolis_mc_dm1_3_6_3_6")
METROPOLIS_MC_DM2_3_6_3_6 = dev_hamiltonian.get_function("metropolis_mc_dm2_3_6_3_6")

METROPOLIS_MC_DM1_6_6_6_12 = dev_hamiltonian.get_function("metropolis_mc_dm1_6_6_6_12")
METROPOLIS_MC_DM0_6_6_6_12 = dev_hamiltonian.get_function("metropolis_mc_dm0_6_6_6_12")

METROPOLIS_MC_DM1_4_4_4_8  = dev_hamiltonian.get_function("metropolis_mc_dm1_4_4_4_8")
METROPOLIS_MC_DM0_4_4_4_8  = dev_hamiltonian.get_function("metropolis_mc_dm0_4_4_4_8")

EN_CALC_3_6_3_6 = dev_hamiltonian.get_function("encalc_3636")
EN_CALC_3_6_3_6_2 = dev_hamiltonian.get_function("encalc_3636_2")
EN_CALC_6_6_6_12 = dev_hamiltonian.get_function("encalc_66612")
EN_CALC_6_6_6_12_2 = dev_hamiltonian.get_function("encalc_66612_2")
EN_CALC_4_4_4_8 = dev_hamiltonian.get_function("encalc_4448")
EN_CALC_4_4_4_8_2 = dev_hamiltonian.get_function("encalc_4448_2")
EN_CALC_2_4_2_4 = dev_hamiltonian.get_function("encalc_2424")
EN_CALC_2_2_4_2 = dev_hamiltonian.get_function("encalc_2242")


import os
from distutils.core import setup
from setuptools import find_packages

# User-friendly description from README.md
current_directory = os.path.dirname(os.path.abspath(__file__))
try:
    with open(os.path.join(current_directory, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
except Exception:
    long_description = ''

setup(
	# Name of the package 
	name='CUDA-METRO',
	version="1.2.1",
	packages=find_packages(),
	# Short description of your library 
	description='pyCUDA Metropolis Monte Carlo 2D Heisenberg Model Simulation',
	# Long description of your library 
	long_description=long_description,
	long_description_content_type='text/markdown',
	# Your name 
	author='Arkavo Hait',
	# Your email 
	author_email='arkavohait@gmail.com',
	# Either the link to your github or to your website 
	url='https://github.com/arkavo/CUDA-METRO',
	# Link from which the project can be downloaded 
	download_url='https://github.com/arkavo/CUDA-METRO/releases/',
	py_modules=["montecarlo"],
	include_package_data=True,
	package_data={'': ['montecarlo.py']},
	# List of keywords 
	keywords=['python',
			   'cuda',
			   'pycuda',
			   'gpu',
			   'parallel computing',
			   'computational physics',
			   ''],
	# List of packages to install with this one 
	install_requires=[	'pycuda==2024.1.2',
						'numpy==2.1.2',
						'seaborn>=0.13.2',
						'tqdm>=4.66.5'],
	# https://pypi.org/classifiers/ 
	classifiers=['Development Status :: 4 - Beta',
	'Environment :: GPU :: NVIDIA CUDA :: 12 :: 12.1' ]
)



import construct as cst
import cudametro.montecarlo as mc
import numpy as np
import tqdm as tqdm
import datetime as dt
import pycuda.autoinit
import pycuda.driver as cuda
import sys

tol = 1e-2


timings, runs = np.array([]), np.array([])
for i in range(6):
    mc_gb = cst.MonteCarlo(config=f"../../configs/grid_benchs/grid_bench_{i}.json")
    mc_gb.mc_init()
    mc_gb.display_material()
    mc_gb.grid_reset()
    print(mc_gb.size, mc_gb.T)
    print(f"Begin Bench at {mc_gb.size} grid")
    t_curr = dt.datetime.now()
    ct = 0
    R_FLAG = False
    while not R_FLAG:
        mc_gb.generate_random_numbers(mc_gb.S_Wrap)
        m,x = mc_gb.run_mc_tc_3636(mc_gb.T[0])
        ct += 1
        if m < tol:
            R_FLAG = True
            pass
    t_end = dt.datetime.now()
    print(f"Time taken for {mc_gb.size} grid {t_end-t_curr}")
    print(f"Number of runs: {ct}")
    timings = np.append(timings, (t_end-t_curr).total_seconds())
    runs = np.append(runs, ct)

np.save(f"Timings", timings)
np.save(f"Runs", runs)

        


import pycuda
import pycuda.autoinit
import pycuda.driver as drv
from pycuda.curandom import rand as curand 

rg = pycuda.curandom.XORWOWRandomNumberGenerator()

import numpy as np 
from numpy import random as rd
import seaborn as sns 
import matplotlib.pyplot as plt 

import os 
import sys 
sys.path.append("../utilities/")
import Material_Reader as rm 
import cudametro.montecarlo as mc
from tqdm import tqdm 

import csv
import json
import datetime

class alt_Montecarlo():
    def __init__(self, config1, config2):
        with open(config1, 'r') as f:
            CONFIG = json.load(f)
        # FLAGS
        self.Single_MAT_Flag = CONFIG["Single_Mat_Flag"]
        self.Animation_Flags = CONFIG["Animation_Flags"]
        self.Static_T_Flag = CONFIG["Static_T_Flag"]
        self.Temps = np.float32(CONFIG["Temps"])
        self.FM_Flag = CONFIG["FM_Flag"]
        self.DMI_Flag = CONFIG["DMI_Flag"]
        self.TC_Flag = CONFIG["TC_Flag"]
        # CONSTANTS AND PATHS
        self.Material = CONFIG["Material"]
        self.Multiple_Materials = CONFIG["Multiple_Materials"]
        self.Input_Folder = "inputs/"
        self.Output_Folder = "outputs/"
        self.B_C = np.float32(CONFIG["B"])
        self.size = CONFIG["SIZE"]
        self.Box = CONFIG["Box"]
        self.Blocks = CONFIG["Blocks"]
        self.Threads = CONFIG["Threads"]
        self.stability_runs = CONFIG["stability_runs"]
        self.calculation_runs = CONFIG["calculation_runs"]
        self.Cmpl = CONFIG["Cmpl_Flag"]
        self.S_Wrap = CONFIG["stability_wrap"]
        self.C_Wrap = CONFIG["calculation_wrap"]
        self.Prefix = CONFIG["Prefix"]
        self.dump_location = "dumps/"
        self.Input_flag = CONFIG["Input_flag"]
        self.Input_File = CONFIG["Input_File"]
        self.C1 = self.Blocks*self.stability_runs
        self.C2 = self.Blocks*self.calculation_runs
        #placeholder spin = 1.0
        self.spin = 1.0
        spin_gpu = np.array([self.spin]).astype(np.float32)
        size_int = np.array([self.size]).astype(np.int32)
        self.b = np.array([self.B_C]).astype(np.float32)
        self.SGPU = drv.mem_alloc(spin_gpu.nbytes)
        self.GSIZE = drv.mem_alloc(size_int.nbytes)
        self.B_GPU = drv.mem_alloc(self.b.nbytes)
        drv.memcpy_htod(self.SGPU, spin_gpu)
        drv.memcpy_htod(self.GSIZE, size_int)
        drv.memcpy_htod(self.B_GPU, self.b)
        self.save_direcotry = "../"+self.Output_Folder+self.Prefix+"_"+self.Material+"_"+str(datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S"))
        os.mkdir(self.save_direcotry)
    
    def generate_random_numbers(self, mult):
        self.NLIST = rg.gen_uniform((self.C1), np.float32)
        self.ULIST = rg.gen_uniform((self.C1), np.float32)
        self.VLIST = rg.gen_uniform((self.C1), np.float32)
        self.RLIST = rg.gen_uniform((self.C1), np.float32)

        self.NFULL = pycuda.gpuarray.zeros((self.C1), dtype=np.int32)
        self.S1FULL = pycuda.gpuarray.zeros((self.C1), dtype=np.float32)
        self.S2FULL = pycuda.gpuarray.zeros((self.C1), dtype=np.float32)
        self.S3FULL = pycuda.gpuarray.zeros((self.C1), dtype=np.float32)

        mc.NPREC(self.NLIST, self.NFULL, self.GSIZE, block=(1,1,1), grid=(self.C1,1,1))
        mc.VPREC(self.ULIST, self.VLIST, self.S1FULL, self.S2FULL, self.S3FULL, self.SGPU, block=(1,1,1), grid=(self.C1,1,1))
    
    def mc_init(self, S1, S2):
        self.grid = np.zeros((self.size*self.size*4), dtype=np.float32)
        self.TMATRIX = np.zeros((self.Blocks, 4)).astype(np.float32)
        self.GPU_TRANS = drv.mem_alloc(self.TMATRIX.nbytes)
        self.MAT_NAME, self.MAT_PARAMS = rm.read_2dmat("../"+self.Input_Folder+"TC_"+self.Material+".csv")
        self.S1 = np.array([S1], dtype=np.float32)
        self.S2 = np.array([S2], dtype=np.float32)
        
        

        self.GPU_MAT = drv.mem_alloc(self.MAT_PARAMS.nbytes)
        self.GRID_GPU = drv.mem_alloc(self.grid.nbytes)
        if self.Static_T_Flag:
            self.T = np.array([self.Temps]).astype(np.float32)
        else:
            self.T = np.linspace(0.01, np.float32(2.0*self.MAT_PARAMS[24]), 7)
        self.BJ = drv.mem_alloc(self.T[0].nbytes)
        drv.memcpy_htod(self.GPU_MAT, self.MAT_PARAMS)
        drv.memcpy_htod(self.GRID_GPU, self.grid)
        DEBUG = np.array([0], dtype=np.int32)
        self.DEBUG_GPU = drv.mem_alloc(DEBUG.nbytes)
        SPINSET = np.array([self.S1, self.S2]).astype(np.float32)
        self.SPINSET_GPU = drv.mem_alloc(SPINSET.nbytes)
        drv.memcpy_htod(self.SPINSET_GPU, SPINSET)
        mc.ALT_GRID(self.GSIZE, self.GRID_GPU, self.DEBUG_GPU, self.SPINSET_GPU, block=(1,1,1),grid=(1,1,1))
        drv.memcpy_dtoh(self.grid, self.GRID_GPU)
        for i in range(self.size*self.size):
            if self.grid[i*4+3] == 1:
                self.grid[i*4+0] *= self.S1[0]
                self.grid[i*4+1] *= self.S1[0]
                self.grid[i*4+2] *= self.S1[0]
            elif self.grid[i*4+3] == 2:
                self.grid[i*4+0] *= self.S2[0]
                self.grid[i*4+1] *= self.S2[0]
                self.grid[i*4+2] *= self.S2[0]

        if self.FM_Flag:
            mc.alt_FM_N(self.grid, S1, S2,self.size)
        else:
            mc.alt_AFM_N(self.grid, S1, S2, self.size)    
        drv.memcpy_htod(self.GRID_GPU, self.grid)
        print(self.GRID_GPU)
        drv.memcpy_dtoh(DEBUG, self.DEBUG_GPU)
        print(self.grid)
        print(DEBUG)
        self.S1GPU = drv.mem_alloc(self.S1.nbytes)
        self.S2GPU = drv.mem_alloc(self.S2.nbytes)
        drv.memcpy_htod(self.S1GPU, self.S1)
        drv.memcpy_htod(self.S2GPU, self.S2)

    def run_mc_3636(self, T):
        beta = np.array([1.0 / (T[0] * 8.6173e-2)],dtype=np.float32)
        drv.memcpy_htod(self.BJ,beta[0])
        for i in range(self.stability_runs):
            mc.METROPOLIS_ALT_MnCr_3_6_3_6(self.GRID_GPU, self.BJ, self.NFULL[i*self.Blocks:(i+1)*self.Blocks-1], self.S1FULL[i*self.Blocks:(i+1)*self.Blocks], self.S2FULL[i*self.Blocks:(i+1)*self.Blocks-1], self.S3FULL[i*self.Blocks:(i+1)*self.Blocks-1], self.RLIST[i*self.Blocks:(i+1)*self.Blocks-1], self.GPU_TRANS,  self.B_GPU, self.GSIZE, self.S1GPU, self.S2GPU, block=(self.Threads,1,1), grid=(self.Blocks,1,1))
            mc.ALT_GRID_COPY(self.GRID_GPU, self.GPU_TRANS, block=(1,1,1), grid=(self.Blocks,1,1))
        drv.memcpy_dtoh(self.grid, self.GRID_GPU)
        return self.grid
    
    def run_mc_tc_3636(self, T):
        Mt, Xt = 0.0, 0.0
        ct = 0
        M = np.zeros(self.S_Wrap)
        for i in tqdm(range(self.S_Wrap), desc=f"Stabilizing at {T:2f}", colour="blue"):
            mag_fluc =  np.zeros(self.stability_runs)
            beta = np.array([1.0 / (T * 8.6173e-2)],dtype=np.float32)
            drv.memcpy_htod(self.BJ, beta)
            for j in range(self.stability_runs):
                mc.METROPOLIS_MALT_MnCr_3_6_3_6(self.GRID_GPU, self.BJ, self.NFULL[j*self.Blocks:(j+1)*self.Blocks-1], self.S1FULL[j*self.Blocks:(j+1)*self.Blocks-1], self.S2FULL[j*self.Blocks:(j+1)*self.Blocks-1], self.S3FULL[j*self.Blocks:(j+1)*self.Blocks-1], self.RLIST[j*self.Blocks:(j+1)*self.Blocks-1], self.GPU_TRANS, self.B_GPU, self.GSIZE, self.S1GPU, self.S2GPU, block=(self.Threads,1,1), grid=(self.Blocks,1,1))
                mc.GRID_COPY(self.GRID_GPU, self.GPU_TRANS, block=(1,1,1), grid=(self.Blocks,1,1))
            drv.memcpy_dtoh(self.grid, self.GRID_GPU)
            self.grid = self.grid.reshape((self.size, self.size, 4))
            magx, magy, magz = self.grid[:,:,0], self.grid[:,:,1], self.grid[:,:,2]
            for k in range(self.size):
                for l in range(self.size):
                    if self.grid[k][l][3] == 1:
                        magx[k][l] /= self.S1[0]
                        magy[k][l] /= self.S1[0]
                        magz[k][l] /= self.S1[0]
                    elif self.grid[k][l][3] == 2:
                        magx[k][l] /= self.S2[0]
                        magy[k][l] /= self.S2[0]
                        magz[k][l] /= self.S2[0]
            mag = np.array([np.sum(magx) , np.sum(magy) , np.sum(magz)])/(self.size**2)
            M[i] = np.abs(np.linalg.norm(mag))
        Mt, Xt = np.mean(M), np.std(M)/T
        print(f"Mean Magnetization at {T} = {Mt}")
        print(f"Mean Susceptibility at {T} = {Xt}")
        return Mt, Xt

    def Analyze(self, reverse=False):
        self.flist = os.listdir(self.save_direcotry)
        self.flist = [file for file in self.flist if file.endswith(".npy")]
        self.flist.sort(reverse=reverse)
        self.directory = self.save_direcotry
        print(self.flist)
    
    def spin_view(self):
        ctr = 0
        mkdir = os.mkdir(self.directory+"/spin_views")
        spin_dir = self.directory+"/spin_views"
        for file in self.flist:
            print(file)
            grid = np.load(self.directory+"/"+file)
            shape = grid.shape
            grid = grid.reshape((int(np.sqrt(shape[0])), int(np.sqrt(shape[0])), 4))
            spinx, spiny, spinz = grid[:,:,0], grid[:,:,1], grid[:,:,2]
            figure = plt.figure(dpi=400)
            plt.title("Spin Configuration at T = "+str(ctr))
            ax = figure.add_subplot(131)
            sns.heatmap(spinz, cbar=False, cmap="coolwarm", square=True, xticklabels=False, yticklabels=False, vmin=-1.0, vmax=1.0)
            ax.set_xlabel("Z")
            ax = figure.add_subplot(132)
            sns.heatmap(spiny, cbar=False, cmap="coolwarm", square=True, xticklabels=False, yticklabels=False)
            ax.set_xlabel("Y")
            ax = figure.add_subplot(133)
            sns.heatmap(spinx, cbar=False, cmap="coolwarm", square=True, xticklabels=False, yticklabels=False)
            ax.set_xlabel("X")
            plt.savefig(f"{spin_dir}/spin_{ctr:04d}.png")
            plt.close()
            ctr += 1

    def quiver_view(self):
        ctr = 0
        mkdir = os.mkdir(self.directory+"/quiver_views")
        spin_dir = self.directory+"/quiver_views"
        for file in self.flist:
            print(file)
            grid = np.load(self.directory+"/"+file)
            shape = grid.shape
            grid = grid.reshape((int(np.sqrt(shape[0])), int(np.sqrt(shape[0])), 4))
            spinx, spiny, spinz = grid[:,:,0], grid[:,:,1], grid[:,:,2]
            x_mesh, y_mesh = np.meshgrid(np.arange(0, int(np.sqrt(shape[0])), 1), np.arange(0, int(np.sqrt(shape[0])), 1))
            figure = plt.figure(dpi=400)
            plt.title("Spin Configuration at T = "+str(ctr))
            ax = figure.add_subplot(111)
            rgba = np.zeros((shape[0],4))
            spinz = np.reshape(spinz, shape[0])/2.0
            for i in range(shape[0]):
                rgba[i][3] = 1.0
                rgba[i][1] = 0.0#np.abs(spinz[i])
                if spinz[i] > 0:
                    rgba[i][0] = spinz[i]
                    rgba[i][2] = 0.0
                else:
                    rgba[i][0] = 0.0
                    rgba[i][2] = -spinz[i]
            plt.quiver(x_mesh, y_mesh, spinx, spiny, scale=1.5, scale_units="xy", pivot="mid", color=rgba, width=0.01, headwidth=3, headlength=4, headaxislength=3, minlength=0.1, minshaft=1)
            plt.savefig(f"{spin_dir}/spin_{ctr:04d}.png")
            plt.close()
            ctr += 1


from cudametro import construct as cst
import numpy as np
import matplotlib.pyplot as plt
import tqdm as tqdm

# =============================================================================
# Edit the following lines to change the parameters of the simulation
# =============================================================================
fname = "fig2_configs/fig2a.json"          # Configuration file [options fig2a.json, fig2b.json]
in_dir = "input_parameters/"               # Input directory
# =============================================================================

# Set your config file here
test_mc0 = cst.MonteCarlo(config=fname, input_folder=in_dir)

if fname.split('/')[-1] == "fig2a.json":

    # Initialize the Monte Carlo simulation
    test_mc0.mc_init()
    test_mc0.display_material()
    for i in tqdm.tqdm(range(test_mc0.S_Wrap), desc="Stability Runs", unit="runs", colour="green"):
        test_mc0.generate_random_numbers(test_mc0.stability_runs)
        np.save(f"{test_mc0.save_directory}/grid_{i:04d}", test_mc0.run_mc_dmi_4448(test_mc0.T[0]))
if fname == "fig2b.json":
    # Initialize the Monte Carlo simulation
    test_mc0.mc_init()
    test_mc0.display_material()
    for i in tqdm.tqdm(range(test_mc0.S_Wrap), desc="Stability Runs", unit="runs", colour="green"):
        test_mc0.generate_random_numbers(test_mc0.stability_runs)
        np.save(f"{test_mc0.save_directory}/grid_{i:04d}", test_mc0.run_mc_dmi_66612(test_mc0.T[0]))
# Save the final state of the simulation 
# To visualize the final state, use the "visualize.py" script with appropriate directory
else:
    print("Invalid config file, options are fig2a.json and fig2b.json")

output_folder_name = test_mc0.save_directory

folder = output_folder_name

viewer = cst.Analyze(folder, reverse=False, input_folder=in_dir)

# Visualize the results
# Spin configuration at the end of the simulation
viewer.spin_view()
# Quiver plot of the spins
viewer.quiver_view()

# Your Diagrams are now saved in the output folder. Use the 'reverse' flag if you want to quickly visualize the end state.
# You can even delete data points should you wish to speed up your visualizer.


# This file contains the function to read the material properties from a file
# Sample file structures are in the 'inputs' folder
#--------------------------------------------
import numpy as np
import re

def read_2dmat(filename):
    namelist = []
    params_list = np.array([])
    with open(filename, 'r') as f:
        data = f.readlines()
        for item in data:
            subdata = re.split(r'[,|+]', item)
            subdata[-1] = subdata[-1].strip("\n")
            namelist.append(subdata[0])
            params_list = np.append(params_list, np.float32(subdata[1:]))
    return namelist, np.array(params_list, dtype=np.float32)


